import threading
import time
import typing as tp
import weakref
from abc import ABCMeta, abstractmethod

from ..heaps import TimeBasedSetHeap
from ..singleton import Singleton
from ...concurrent.monitor import Monitor
from ...recast_exceptions import rethrow_as, silence_excs

K, V = tp.TypeVar('K'), tp.TypeVar('V')


class Cleanupable(metaclass=ABCMeta):
    @abstractmethod
    def cleanup(self) -> None:
        ...


@Singleton
class ExpiringEntryDictThread(threading.Thread, Monitor):
    """
    A background thread providing maintenance for expiring entry dicts
    and self-cleaning default dicts
    """

    def __init__(self):
        super().__init__(name='ExpiringEntryDict cleanup thread', daemon=True)
        Monitor.__init__(self)
        self.entries = []  # type: tp.List[weakref.ref[Cleanupable]]
        self.started = False  # type: bool

    def start(self) -> None:
        if self.started:
            return

        self.started = True
        super().start()

    def run(self) -> None:
        while True:
            time.sleep(5)
            self.cleanup()

    @Monitor.synchronized
    def cleanup(self) -> None:
        new_entries = []
        for index, ref in enumerate(self.entries):
            obj = ref()
            if obj is not None:
                obj.cleanup()
                new_entries.append(ref)
        self.entries = new_entries

    @Monitor.synchronized
    def add_dict(self, ed: Cleanupable) -> None:
        self.entries.append(weakref.ref(ed))
        self.start()


class SelfCleaningDefaultDict(Monitor, tp.MutableMapping[K, V], Cleanupable):
    """
    A defaultdict with the property that if it detects that a value is equal to it's default value,
    it will automatically remove it from the dict.

    Please note that this will spawn a cleanup thread in the background, one per program. The thread
    is shared between :class:`satella.coding.structures.SelfCleaningDefaultDict` and
    :class:`satella.coding.structures.ExpiringEntryDict`

    It is preferable to :meth:`satella.coding.concurrent.Monitor.acquire` it before iterating, since
    it is cleaned up both by __iter__ and by an external worker thread, so it's important to acquire
    it, because it can mutate in an undefined way.

    Note that if you access a key which does not exist, and background_maintenance is False,
    a default value will be created and inserted into the dictionary. This is the only
    time that the dictionary will hold values that are equal to default.

    :param default_factory: a callable/0 that will return an object if it doesn't already exist
    :param background_maintenance: whether to spawn a background thread. This is required
        if dictionary values can change their value between inserts.

    All args and kwargs will be passed to a dict, which will be promptly added to this dictionary.
    """

    def __len__(self) -> int:
        self.cleanup()
        return len(self.data)

    def __init__(self, default_factory: tp.Callable[[], V],
                 background_maintenance: bool = True, *args, **kwargs):
        super().__init__()  # initialize the inner Monitor
        self.data = dict(*args, **kwargs)
        self.default_factory = default_factory
        self.default_value = default_factory()

        self.background_maintenance = background_maintenance
        if self.background_maintenance:
            ExpiringEntryDictThread().add_dict(self)

    def __iter__(self) -> tp.Iterator[K]:
        self.cleanup()
        return super().__iter__()

    def __delitem__(self, key: K) -> None:
        if key in self.data:
            del self.data[key]

    @Monitor.synchronized
    def __getitem__(self, item: K) -> V:
        try:
            v = self.data[item]
            if v == self.default_value:
                del self.data[item]
            return v
        except KeyError:
            obj = self.default_factory()
            if not self.background_maintenance:
                self.data[item] = obj
            return obj

    @Monitor.synchronized
    def __setitem__(self, key: K, value: V) -> None:
        value_equal = value == self.default_value
        if key in self.data:
            if value_equal:
                del self.data[key]

        if not value_equal:
            self.data[key] = value

    @Monitor.synchronized
    @silence_excs(KeyError)  # because entries may disappear without warning
    def cleanup(self) -> None:
        for key in list(self.data.keys()):
            if self.data[key] == self.default_value:
                del self.data[key]


class ExpiringEntryDict(Monitor, tp.MutableMapping[K, V], Cleanupable):
    """
    A dictionary whose entries expire automatically after a predefined period of time.

    Note that cleanup is invoked only when iterating over the dicts, or automatically if you specify
    external_cleanup to be True each approximately 5 seconds.

    Note that it's preferential to :meth:`satella.coding.concurrent.Monitor.acquire` it if you're
    using an external cleanup thread, because the dict may mutate at any time.

    :param expiration_timeout: number of seconds after which entries will expire
    :param time_getter: a callable/0 that returns the current timestamp
    :param external_cleanup: whether to spawn a single thread that will clean up the dictionary.
        The thread is spawned once per program, and no additional threads are spawned for next
        dictionaries.

    All args and kwargs will be passed to a dict, which will be promptly added to this dictionary.
    """

    def __len__(self) -> int:
        self.cleanup()
        return len(self.data)

    def __init__(self, expiration_timeout: float, *args,
                 time_getter: tp.Callable[[], float] = time.monotonic,
                 external_cleanup: bool = False, **kwargs):
        super().__init__()  # initialize the inner Monitor
        self.data = dict()
        self.expire_on = {}
        self.time_getter = time_getter
        self.expiration_timeout = expiration_timeout
        self.key_to_expiration_time = TimeBasedSetHeap()

        if external_cleanup:
            ExpiringEntryDictThread().add_dict(self)

        dct = dict(*args, **kwargs)
        for key, value in dct.items():
            self[key] = value

    def __contains__(self, item) -> bool:
        try:
            ts = self.key_to_expiration_time.item_to_timestamp[item]
        except KeyError:
            return False

        if ts < self.time_getter():
            del self[item]
            return False

        return True

    def get_timestamp(self, key: K) -> float:
        """
        Return the timestamp at which given key was inserted in the dict

        :raises KeyError: key not found in the dictionary
        """
        return self.key_to_expiration_time.item_to_timestamp[key]

    def __iter__(self) -> tp.Iterator[K]:
        self.cleanup()
        return super().__iter__(self)

    @Monitor.synchronized
    def cleanup(self) -> None:
        """Remove entries that are later than given time"""
        for ts, key in self.key_to_expiration_time.pop_less_than(self.time_getter()):
            del self.data[key]

    @Monitor.synchronized
    def __setitem__(self, key: K, value: V) -> None:
        self.key_to_expiration_time.put(self.time_getter() + self.expiration_timeout, key)
        self.data[key] = value

    @rethrow_as(ValueError, KeyError)
    def __getitem__(self, item: K) -> V:
        ts = self.key_to_expiration_time.get_timestamp(item)
        if ts < self.time_getter():
            del self.data[item]
            raise KeyError('Entry expired')

        return self.data[item]

    @Monitor.synchronized
    def __delitem__(self, key: K) -> None:
        self.key_to_expiration_time.pop_item(key)
        del self.data[key]



import typing as tp

from satella.coding.concurrent import CallableGroup
from satella.exceptions import ConfigurationValidationError
from .registry import register_custom_descriptor

ConfigDictValue = tp.Optional[tp.Union[int, float, str, dict, list, bool]]
CheckerConditionType = tp.Callable[[ConfigDictValue], bool]
ObjectMakerType = tp.Callable[
    [ConfigDictValue], tp.Any]  # might raise ConfigurationSchemaError as well


class CheckerCondition:
    __slots__ = ('condition', 'description', 'is_pre_checker')

    PRE_CHECKER = 0
    POST_CHECKER = 1

    def __init__(self, condition: CheckerConditionType, description: str = u'',
                 is_pre_checker: bool = True):
        self.condition = condition
        self.description = description
        self.is_pre_checker = is_pre_checker

    def __call__(self, value):
        if not self.condition(value):
            raise ConfigurationValidationError(self.description, value)


def must_be_type(*cls_or_tuple):
    return CheckerCondition(condition=lambda v: isinstance(v, cls_or_tuple),
                            description='not one of types %s' % (cls_or_tuple,))


def must_be_one_of(*items):
    return CheckerCondition(condition=lambda v: v in items,
                            description='not in set %s' % (items,),
                            is_pre_checker=False)


@register_custom_descriptor('any')
class Descriptor:
    """
    Base class for a descriptor
    """
    __slots__ = ('pre_checkers', 'post_checkers', 'name', 'optional', 'default',
                 'my_exceptions')

    BASIC_MAKER = staticmethod(lambda v: v)
    MY_EXCEPTIONS = [TypeError, ValueError]  # a list of Exception classes
    CHECKERS = []  # a list of CheckerCondition

    def __init__(self):
        self.pre_checkers = CallableGroup()  # type: tp.Callable[[bool], None]
        self.post_checkers = CallableGroup()  # type: tp.Callable[[bool], None]
        self.name = None  # type: tp.Optional[str]
        self.optional = None  # type: tp.Optional[bool]
        self.default = None  # type: tp.Optional[tp.Any]

        for checker in self.__class__.CHECKERS:
            self.add_checker(checker)

        self.my_exceptions = tuple(self.MY_EXCEPTIONS)  # type: tp.Tuple[tp.Type[Exception], ...]

    def __str__(self):
        return '%s()' % (self.__class__.__qualname__,)

    def __call__(self, value: ConfigDictValue) -> tp.Any:
        """
        raises ConfigurationSchemaError: on invalid schema
        """
        self.pre_checkers(value)

        try:
            value = self.BASIC_MAKER(value)
        except self.my_exceptions as e:
            raise ConfigurationValidationError('could not pass to maker due to %s' % (e,), value)

        self.post_checkers(value)

        return value

    def add_checker(self, checker: CheckerCondition):
        if checker.is_pre_checker:
            self.pre_checkers.add(checker)
        else:
            self.post_checkers.add(checker)



import typing as tp

from satella.coding.recast_exceptions import silence_excs


def chain(*args) -> tp.Iterator:
    """
    Construct an iterator out of provided elements.

    If an element is an iterator, or an iterable it will be yielded-from. If it's not, it will
    just be yielded.

    A cast to iter() is used to determine iteratorness
    """
    for elem in args:
        try:
            yield from iter(elem)
        except TypeError:
            yield elem


def exhaust(iterator: tp.Iterator) -> None:
    """
    Iterate till the end of the iterator, discarding values as they go

    :param iterator: iterator to exhaust
    """
    try:
        while True:
            next(iterator)
    except StopIteration:
        pass


class SelfClosingGenerator:
    """
    A wrapper to exhaust the generator in response to closing it.

    This will allow generators to complete that don't provide a .close() method.

    This will additionally exhaust the generator upon deallocation of the generator.

    You can feed it with either generators, or generator-functions, it will behave correctly each
    time.

    You can also use it as a context manager, to decouple finalizing the generator from the GC
    collection
    """
    __slots__ = ('generator', 'stopped')

    def __init__(self, generator: tp.Generator):
        self.generator = generator
        self.stopped = False  # type: bool

    def __iter__(self) -> 'SelfClosingGenerator':
        return self

    def __call__(self, *args, **kwargs):
        return SelfClosingGenerator(self.generator(*args, **kwargs))

    def send(self, obj: tp.Any) -> None:
        self.generator.send(obj)

    def __next__(self):
        try:
            return next(self.generator)
        except StopIteration:
            self.stopped = True
            raise

    def __enter__(self) -> 'SelfClosingGenerator':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        self.close()
        return False

    def close(self) -> None:
        if not self.stopped:
            with silence_excs(TypeError):  # we might get a generator-generating function
                                           # as an argument
                exhaust(self.generator)

            self.stopped = True

    def __del__(self):
        self.close()


# noinspection PyPep8Naming
class hint_with_length:
    """
    Accepting a generator, return it additionally providing
    a specified __length_hint__

    You can provide generator-generating functions as well

    :param generator: generator to decorate
    :param length: length hint to provide
    :param length_factory: a callable called with no arguments to get the length

    You must provide either length or length_factory. Giving them both is wrong, and
    will result in ValueError
    """
    __slots__ = ('generator', 'length', 'length_factory')

    def __init__(self, generator: tp.Generator, length: tp.Optional[int],
                 length_factory: tp.Optional[tp.Callable[[], int]] = None):
        self.generator = generator
        if length is None and length_factory is None:
            raise ValueError('Give either length or length_factory')
        elif length is not None and length_factory is not None:
            raise ValueError('Give either length or length_factory')

        self.length = length
        self.length_factory = length_factory

    def __call__(self, *args, **kwargs):
        return hint_with_length(self.generator(*args, **kwargs), self.length)

    def close(self):
        return self.generator.close()

    def send(self, obj):
        return self.generator.send(obj)

    def __iter__(self) -> tp.Generator:
        return self.generator

    def __next__(self):
        return next(self.generator)

    def __length_hint__(self) -> int:
        if self.length is None:
            return self.length_factory()
        else:
            return self.length



import itertools
import collections
import typing as tp
import warnings

from ..recast_exceptions import rethrow_as, silence_excs
from ..decorators import for_argument, wraps

T, U = tp.TypeVar('T'), tp.TypeVar('U')
IteratorOrIterable = tp.Union[tp.Iterator[T], tp.Iterable[T]]


def length(iterator: IteratorOrIterable) -> int:
    """
    Return the length of an iterator, exhausting it by the way
    """
    i = 0
    for _ in iterator:
        i += 1
    return i


def walk(obj: T, child_getter: tp.Callable[[T], tp.Optional[tp.List[T]]] = list,
         deep_first: bool = True,
         leaves_only: bool = False) -> tp.Iterator[T]:
    """
    Return every node of a nested structure.

    :param obj: structure to traverse. This will not appear in generator
    :param child_getter: a callable to return a list of children of T.
        Should return an empty list or None of there are no more children.
    :param deep_first: if True, deep first will be returned, else it will be breadth first
    :param leaves_only: if True, only leaf nodes (having no children) will be returned
    """
    a = ConstruableIterator(child_getter(obj))
    for o in a:
        children = child_getter(o)
        if children is not None:
            try:
                child_len = len(children)
            except TypeError:
                child_len = 0

            if child_len:
                if deep_first:
                    a.add_many_immediate(children)
                else:
                    a.add_many(children)
                if leaves_only:
                    continue

        yield o


class ConstruableIterator:
    """
    An iterator that you can attach arbitrary things at the end and consume them during iteration.
    Eg:

    >>> a = ConstruableIterator([1, 2, 3])
    >>> for b in a:
    >>>   if b % 2 == 0:
    >>>       a.add(6)

    All arguments you provide to the constructor will be passed to underlying deque
    """
    __slots__ = ('entries', )

    def __init__(self, *args, **kwargs):
        self.entries = collections.deque(*args, **kwargs)       # type: tp.List[T]

    def add_immediate(self, t: T) -> None:
        """
        Schedule given value to be iterated over during the next __next__ call

        :param t: value to iterate over
        """
        self.entries.appendleft(t)

    def add_many_immediate(self, t: tp.Iterable[T]) -> None:
        """
        Schedule given values to be iterated over during the next __next__ call

        :param t: values to iterate over
        """
        for i, entry in enumerate(t):
            self.entries.insert(i, entry)

    def add(self, t: T) -> None:
        """
        Schedule given value to be iterated over after current items

        :param t: value to iterate over
        """
        self.entries.append(t)

    def add_many(self, t: tp.Iterable[T]) -> None:
        """
        Schedule given values to be iterated over after current items

        :param t: iterable of values
        """
        self.entries.extend(t)

    def __iter__(self) -> 'ConstruableIterator':
        return self

    @rethrow_as(IndexError, StopIteration)
    def __next__(self) -> T:
        return self.entries.popleft()

    def __length_hint__(self) -> int:
        return len(self.entries)


def unique(lst: IteratorOrIterable) -> tp.Iterator[T]:
    """
    Return each element from lst, but return every element only once.

    Take care for elements of T to be __eq__-able and hashable!

    This will keep internally a set of elements encountered, and skip them if same element
    appears twice

    :param lst: iterable to process
    :return: a generator yielding unique items from lst
    """
    already_seen = set()
    for elem in lst:
        if elem not in already_seen:
            already_seen.add(elem)
            yield elem


@for_argument(iter)
def even(sq: IteratorOrIterable) -> tp.Iterator[T]:
    """
    Return only elements with even indices in this iterable (first element will be returned,
    as indices are counted from 0)
    """
    while True:
        try:
            yield next(sq)
            next(sq)
        except StopIteration:
            return


@for_argument(iter)
@silence_excs(StopIteration)
def odd(sq: IteratorOrIterable) -> tp.Iterator[T]:
    """
    Return only elements with odd indices in this iterable.
    """
    while True:
        try:
            next(sq)
            yield next(sq)
        except StopIteration:
            return


def count(sq: IteratorOrIterable, start: tp.Optional[int] = None, step: int = 1,
          start_at: tp.Optional[int] = None) -> tp.Iterator[int]:
    """
    Return a sequence of integers, for each entry in the sequence with provided step.

    Essentially the same (if step were ignored) as:

    >>> (i for i, x in enumerate(sq, start=start_at))

    :param sq: sequence to enumerate
    :param start: alias for start_at. Prefer it in regards to start_at. Default is 0
    :param step: number to add to internal counter after each element
    :return: an iterator of subsequent integers
    """
    if start_at:
        start = start_at
    num = start or 0
    for _ in sq:
        yield num
        num += step


def iter_dict_of_list(dct: tp.Dict[T, tp.List[U]]) -> tp.Generator[tp.Tuple[T, U], None, None]:
    """
    Presents a simple way to iterate over a dictionary whose values are lists.

    This will return the dictionary key and each of the value contained in the list attached to
    the key.
    """
    for key, items in dct.items():
        for item in items:
            yield key, item


def infinite_counter(start_at: int = 0, step: int = 1) -> tp.Iterator[int]:
    """
    Infinite counter, starting at start_at

    :param start_at: value at which to start counting. It will be yielded as first
    :param step: step by which to progress the counter
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0, '
                  'use itertools.count() instead', DeprecationWarning)
    i = start_at
    while True:
        yield i
        i += step


def is_instance(classes: tp.Union[tp.Tuple[type, ...], type]) -> tp.Callable[[object], bool]:
    def inner(object_):
        return isinstance(object_, classes)

    inner.__doc__ = """Return a bool telling if object is of type %s""" % (repr(classes),)
    return inner


@for_argument(iter, iter)
def other_sequence_no_longer_than(base_sequence: IteratorOrIterable,
                                  other_sequence: IteratorOrIterable) -> tp.Iterator[T]:
    """
    Return every item in other_sequence, but limit it's p_len to that of base_sequence.

    If other_sequence is shorter than base_sequence, the shorter one will be returned.

    :param base_sequence: sequence whose p_len should be taken
    :param other_sequence: sequence to output values from
    """
    while True:
        try:
            next(base_sequence)
            yield next(other_sequence)
        except StopIteration:
            return


def shift(iterable_: IteratorOrIterable, shift_factor: int) -> tp.Iterator[T]:
    """
    Return this sequence, but shifted by factor elements, so that elements will appear
    sooner by factor.

    Eg:

    >>> assert list(shift([1,2, 3], 1)) == [2, 3, 1]

    However note that this will result in iterators which have negative shift to be readed entirely
    into memory (converted internally to lists). This can be avoided by passing in a Reversible
    iterable.

    :param iterable_: iterable to shift
    :param shift_factor: factor by which shift elements.
    :return: shifted sequence
    """

    if shift_factor >= 0:
        iterator = iter(iterable_)
        elements = []
        for i in range(shift_factor):
            elements.append(next(iterator))
        return itertools.chain(iterator, elements)
    else:
        if hasattr(iterable_, '__reversed__'):
            elements = take_n(reversed(iterable_), -shift_factor)
            elements = reversed(elements)
            return other_sequence_no_longer_than(iterable_, itertools.chain(elements, iterable_))
        else:
            iterator = list(iterable_)
            iterator = iterator[shift_factor:] + iterator[:shift_factor]  # shift's already negative
            return iterator


@silence_excs(StopIteration)
def zip_shifted(*args: tp.Union[IteratorOrIterable, tp.Tuple[IteratorOrIterable, int]]) -> \
        tp.Iterator[tp.Tuple[T, ...]]:
    """
    Construct an iterator, just like zip but first by cycling it's elements by it's shift factor.
    Elements will be shifted by a certain factor, this means that they will appear earlier.

    Example:

    >>> zip_shifted(([1, 2, 3, 4], 1), ([1, 2, 3, 4], 0)) == [(2, 1), (3, 2), (4, 3), (1, 4)]

    This will work on arbitrary iterators and iterables.

    Shift can be negative, in which case the last elements will appear sooner, eg.

    >>> zip_shifted(([1, 2, 3, 4], -1), ([1, 2, 3, 4], 0)) == [(4, 1), (1, 2), (2, 3), (3, 4)]

    Same memory considerations as :func:`shift` apply.

    The resulting iterator will be as long as the shortest sequence.
    :param args: a tuple with the iterator/iterable and amount of shift. If a non-tuple is given,
        it is assumed that the shift is zero.
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0. '
                  'Use zip(shift(...)) instead!', DeprecationWarning)

    iterators = []  # type: tp.List[tp.Union[tp.Tuple[tp.Iterator[T], tp.List[T]], tp.Iterator[T]]
    for row in args:
        if not isinstance(row, tuple):
            iterators.append(row)
        else:
            iterable, shift_factor = row
            iterators.append(shift(iterable, shift_factor))
    return zip(*iterators)


@for_argument(iter)
@silence_excs(StopIteration)
def skip_first(iterator: IteratorOrIterable, n: int) -> tp.Iterator[T]:
    """
    Skip first n elements from given iterator.

    Returned iterator may be empty, if source iterator is shorter or equal to n.
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0. '
                  'Please use itertools.islice instead', DeprecationWarning)

    for i in range(n):
        next(iterator)
    yield from iterator


@for_argument(iter)
@silence_excs(StopIteration)
def stop_after(iterator: IteratorOrIterable, n: int) -> tp.Iterator[T]:
    """
    Stop this iterator after returning n elements, even if it's longer than that.

    The resulting iterator may be shorter than n, if the source element is so.

    :param iterator: iterator or iterable to examine
    :param n: elements to return
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0. '
                  'Please use itertools.islice instead', DeprecationWarning)

    for i in range(n):
        yield next(iterator)


@for_argument(iter)
def n_th(iterator: IteratorOrIterable, n: int = 0) -> T:
    """
    Obtain n-th element (counting from 0) of an iterable

    :param iterator: iterable to process
    :param n: element to return. Note that we're counting from 0
    :raises IndexError: iterable was too short
    """
    try:
        for i in range(n):
            next(iterator)
        return next(iterator)
    except (StopIteration, GeneratorExit):
        raise IndexError('Iterable was too short')


class IteratorListAdapter:
    """
    A wrapper around an iterator that enables it to be processed as a list.

    Ie. the generator will now support __contains__, __len__ and __getitem__.
    If a call to such a method is done, the generator will be unfolded in memory so this
    might take a ton of memory! You've been warned!

    :param iterator:
    """
    __slots__ = ('unfolded', 'iterator', 'list', 'pointer')

    def __init__(self, iterator: tp.Iterator):
        self.iterator = iter(iterator)
        self.list = None        # type: list
        self.unfolded = False
        self.pointer = 0

    def __iter__(self):
        if self.unfolded:
            return iter(self.list)
        else:
            return self

    @rethrow_as(IndexError, StopIteration)
    def __next__(self):
        if self.unfolded:
            self.pointer += 1
            return self.list[self.pointer-1]    # throws: IndexError
        else:
            return next(self.iterator)

    def __unfold(self):
        if self.unfolded:
            return
        self.list = list(self.iterator)
        self.unfolded = True
        self.pointer = 0

    def __getitem__(self, item):
        self.__unfold()
        return self.list[item]

    def __contains__(self, item) -> bool:
        self.__unfold()
        return item in self.list

    def __len__(self) -> int:
        self.__unfold()
        return len(self.list)


def is_empty(iterable: IteratorOrIterable) -> bool:
    """
    Checks whether an iterator is empty.

    This will exhaust the iterator.

    :param iterable: iterator to check
    :return: whether the iterator is empty
    """
    i = 0
    for elem in iterable:
        i += 1
    return i == 0


def map_list(fun: tp.Callable, iterable: IteratorOrIterable) -> tp.List:
    """
    A syntactic sugar for

    >>> list(map(fun, iterable))

    :param fun: function to apply
    :param iterable: iterable to iterate over
    """
    return list(map(fun, iterable))


def to_iterator(fun):
    """
    Convert function to an iterator. You can replace the following code:

    >>> def iterator(x):
    >>>     for y in x:
    >>>         yield fun(y)

    with

    >>> @to_iterator
    >>> def fun(y):
    >>>     ...

    and now call fun instead of iterator. fun will accept a single argument - the iterable,
    and assume that the function you decorate also takes a single argument - the item
    """
    @wraps(fun)
    def inner(iterable):
        for item in iterable:
            yield fun(item)
    return inner


def smart_zip(*iterators: IteratorOrIterable) -> tp.Iterator[tp.Tuple[T, ...]]:
    """
    Zip in such a way that resulted tuples are automatically expanded.

    Ie:

    >>> b = list(smart_zip([(1, 1), (1, 2)], [1, 2]))
    >>> assert b == [(1, 1, 1), (1, 2, 2)]

    Note that an element of the zipped iterator must be a tuple (ie. isinstance tuple)
    in order for it to be appended to resulting iterator element!

    :param iterators: list of iterators to zip together
    :return: an iterator zipping the arguments in a smart way
    """
    for row in zip(*iterators):
        a = []
        for elem in row:
            if isinstance(elem, tuple):
                a.extend(elem)
            else:
                a.append(elem)
        yield tuple(a)


def enumerate2(iterable: IteratorOrIterable, start: int = 0,
               step: int = 1) -> tp.Iterator[tp.Tuple[int, T]]:
    """
    Enumerate with a custom step

    :param iterable: iterable to enumerate
    :param start: value to start at
    :param step: step to add during each iteration
    """
    v = start
    for item in iterable:
        yield v, item
        v += step


def smart_enumerate(iterator: IteratorOrIterable, start: int = 0,
                    step: int = 1) -> tp.Iterator[tp.Tuple]:
    """
    An enumerate that talks pretty with lists of tuples. Consider

    >>> a = [(1, 2), (3, 4), (5, 6)]
    >>> for i, b in enumerate(a):
    >>>     c, d = b
    >>>     ...

    This function allows you just to write:
    >>> for i, c, d in enumerate(a):
    >>>     ...

    Note that elements in your iterable must be either a list of a tuple for that to work,
    or need to be able to be coerced to a tuple. Otherwise, TypeError will be thrown.

    :param iterator: iterator to enumerate
    :param start: value to start counting at
    :param step: step to advance the enumeration with
    :raise TypeError: could not coerce the elements in your iterable to a tuple
    """
    i = start
    for row in iterator:
        if isinstance(row, tuple):
            yield (i,) + row
        else:
            yield (i,) + tuple(row)
        i += step


@for_argument(iter)
def take_n(iterator: IteratorOrIterable, n: int, skip: int = 0) -> tp.List[T]:
    """
    Take (first) n elements of an iterator, or the entire iterator, whichever comes first

    :param iterator: iterator to take from
    :param n: amount of elements to take
    :param skip: elements from the start to skip
    :return: list of p_len n (or shorter)
    """
    for i in range(skip):
        next(iterator)

    output = []
    for i in range(n):
        try:
            output.append(next(iterator))
        except StopIteration:
            return output
    return output



import copy
import typing as tp
import collections

ITER_KEYS = 0
ITER_VALUES = 1
ITER_ITEMS = 2
T = tp.TypeVar('T')


class DictDeleter:
    """
    Having problems deleting entries from your dict while iterating on them? No problem. Just swap
    the following:

    >>> keys_to_delete = []
    >>> for key, value in my_dict.items():
    >>>     if value.should_delete():
    >>>         keys_to_delete.append(key)
    >>> for key in keys_to_delete:
    >>>     del my_dict[key]

    With the following:

    >>> with DictDeleter(my_list) as ld:
    >>>     for key, value in ld.items():
    >>>         if value.should_delete():
    >>>             ld.delete()

    Note that a single DictDeleter running from a single context must be iterated on by only a
    single Thread as it keeps the state of iterator in itself, to prevent allocating new objects
    and slowing things down too much.

    This allocates only a single object per a call to delete().
    """
    __slots__ = ('dict_to_process', 'current_iterator', 'keys_to_delete', 'iter_mode',
                 'current_key')

    def __init__(self, dict_to_process: collections.abc.MutableMapping):
        self.dict_to_process = dict_to_process
        self.iter_mode = ITER_KEYS

    def __enter__(self):
        self.keys_to_delete = set()
        return self

    def __iter__(self):
        self.current_iterator = iter(self.dict_to_process.items())
        return self

    def __next__(self):
        key, value = next(self.current_iterator)  # raises StopIteration
        self.current_key = key
        if self.iter_mode == ITER_ITEMS:
            return key, value
        elif self.iter_mode == ITER_VALUES:
            return value
        elif self.iter_mode == ITER_KEYS:
            return key

    def items(self) -> 'DictDeleter':
        self.iter_mode = ITER_ITEMS
        return self

    def keys(self) -> 'DictDeleter':
        self.iter_mode = ITER_KEYS
        return self

    def values(self) -> 'DictDeleter':
        self.iter_mode = ITER_VALUES
        return self

    def delete(self) -> None:
        self.keys_to_delete.add(self.current_key)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for key in self.keys_to_delete:
            del self.dict_to_process[key]
        return False


DIR_FORWARD = 0
DIR_BACKWARD = 1


class ListDeleter(tp.Generic[T]):
    """
    Having problems deleting entries from your list while iterating on them? No problem. Just swap
    the following:

    >>> entries_to_delete = []
    >>> for entry in my_list:
    >>>     if entry.should_delete():
    >>>         entries_to_delete.append(entry)
    >>> for entry in entries_to_delete:
    >>>     my_list.remove(entry)

    With the following:

    >>> with ListDeleter(my_list) as ld:
    >>>     for entry in ld:
    >>>         if entry.should_delete():
    >>>             ld.delete()

    Note that a single ListDeleter running from a single context must be iterated on by only a
    single Thread as it keeps the state of iterator in itself, to prevent allocating new objects
    and slowing things down too much.

    This allocates only a single object per a call to delete().

    Calling the list deleter during iteration will yield the element.

    You can pass any type of object here, as long as it supports pop(position) and __getitem__
    """
    __slots__ = ('list_to_process', 'current_index', 'indices_to_delete', 'direction')

    def __init__(self, list_to_process: tp.MutableSequence[T]):
        self.list_to_process = list_to_process
        self.direction = DIR_FORWARD
        # pointer to currently processed element
        self.current_index = -1 if self.direction == DIR_FORWARD else len(self.list_to_process)
        self.indices_to_delete = set()      # type: tp.Set[int]

    def __enter__(self) -> 'ListDeleter':
        return self

    @property
    def value(self) -> T:
        return self()

    def __call__(self) -> T:
        """
        Return current element
        """
        return self.list_to_process[self.current_index]

    def __iter__(self) -> 'ListDeleter':
        self.current_index = -1 if self.direction == DIR_FORWARD else len(self.list_to_process)
        self.indices_to_delete = set()
        return self

    def __next__(self):
        if self.direction == DIR_BACKWARD and self.current_index == 0:
            raise StopIteration('First element reached')
        if self.direction == DIR_FORWARD and self.current_index == len(self.list_to_process) - 1:
            raise StopIteration('Last element reached')
        self.current_index += +1 if self.direction == DIR_FORWARD else -1
        return self.list_to_process[self.current_index]

    def next(self) -> T:
        """
        :return: the next element
        :raises StopIteration: no more entries
        """
        next(self)
        return self.list_to_process[self.current_index]

    def prev(self) -> T:
        """
        Move to previous element

        :return: the previous element
        :raises ValueError: list is already at the first element!
        """
        if self.direction == DIR_FORWARD and self.current_index == 0:
            raise ValueError('Cannot go previous on a first element!')
        if self.direction == DIR_BACKWARD and self.current_index == len(self.list_to_process) - 1:
            raise ValueError('Cannot go previous on a last element!')
        self.current_index += -1 if self.direction == DIR_FORWARD else +1
        return self.list_to_process[self.current_index]

    def delete(self) -> None:
        self.indices_to_delete.add(self.current_index)

    def __reversed__(self) -> 'ListDeleter':
        """
        Take care, for this will modify the current object.

        This is done not to break the context manager semantics, and enable this:
        >>> with ListDeleter(a):
        >>>     for v in reversed(a):
        >>>         ...
        """
        if self.direction == DIR_BACKWARD:
            self.direction = DIR_FORWARD
            self.current_index = -1
        else:
            self.direction = DIR_BACKWARD
            self.current_index = len(self.list_to_process)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        index_delta = 0
        for index, element in enumerate(self.list_to_process[:]):
            if index in self.indices_to_delete:
                self.list_to_process.pop(index - index_delta)
                index_delta += 1

        return False

    def __copy__(self) -> 'ListDeleter':
        ld = ListDeleter(self.list_to_process)
        ld.direction = self.direction
        ld.current_index = self.current_index
        ld.indices_to_delete = copy.copy(self.indices_to_delete)
        return ld



import time
from concurrent.futures import wait, ThreadPoolExecutor

from .atomic import AtomicNumber
from .thread import Condition


def sync_threadpool(tp: ThreadPoolExecutor) -> None:
    """
    Make sure that every thread of given thread pool executor is done processing
    jobs scheduled until this moment.

    Make sure that other tasks do not submit anything to this thread pool executor.

    :param tp: thread pool executor to sync
    """
    assert isinstance(tp, ThreadPoolExecutor), 'Must be a ThreadPoolExecutor!'

    workers = tp._max_workers
    atm_n = AtomicNumber(workers)
    cond = Condition()

    def decrease_atm():
        nonlocal atm_n
        atm_n -= 1
        cond.wait()

    futures = [tp.submit(decrease_atm) for _ in range(workers)]

    # wait for all currently scheduled jobs to be picked up
    while tp._work_queue.qsize() > 0:
        time.sleep(0.5)

    atm_n.wait_until_equal(0)
    cond.notify_all()
    wait(futures)




import random
import typing as tp

__all__ = ['shuffle_together']


def shuffle_together(*args: tp.Sequence) -> tp.List[tp.List]:
    """
    args, being sequences of equal length, will be permuted in such a way
    that their indices will still correspond to each other.

    So given:

    >>> a = [1, 2, 3]
    >>> b = ['a', 'b', 'c']
    >>> c = shuffle_together(a, b)

    Might equal

    >>> c == [[3, 1, 2], ['c', 'a', 'b']]

    """
    try:
        indices = list(range(len(args[0])))
    except IndexError:
        return []  # empty array

    random.shuffle(indices)
    return [[arg[i] for i in indices] for arg in args]
